#include <stdlib.h>

// This should be (but doesn't seem to be in VCC South) defined by math.h/cmath
#ifndef M_PI
#define M_PI 3.1415926535
#endif

#include <GL/glut.h>
#include <cmath>
#include <iostream>

int gWidth;
int gHeight;
bool gKeys[256];
const int gDelay = 15;
float gCameraPos[3];
float gCameraRot = 0.0f;
float gCameraRotX = 0.0f;
float gVelocityY;
bool gButtons[3];
int gCursor[2];
float gWindmill;
float gWobble;
GLUquadric *gQuadric;
struct {
	bool outline;
	bool lighting;
	bool movement;
} gOptions;

void reset() {
	gCameraPos[0] = 0.0f;
	gCameraPos[1] = 1.0f;
	gCameraPos[2] = 5.0f;
	gCameraRot = 0.0f;
}

void init() {
	glClearColor(0.0f, 0.0f, 0.7f, 0.0f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	gQuadric = gluNewQuadric();
	//glEnable(GL_LIGHTING);
	//glEnable(GL_LIGHT0);
	reset();
}

void drawHill(float x, float y, float z, float base, float height) {
	glPushMatrix();
	glTranslatef(x, y, z);
	glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
	gluCylinder(gQuadric, base, 0.1f, height, 12, 1);
	glScalef(-1.0f, 1.0f, 1.0f);
	gluDisk(gQuadric, 0.0f, base, 12, 1);
	glScalef(-1.0f, 1.0f, 1.0f);
	glTranslatef(0.0f, 0.0f, height);
	gluDisk(gQuadric, 0.0f, 0.1f, 12, 1);
	glPopMatrix();
}

void drawScene(bool color) {
		float brown[] = {0.5f, 0.3f, 0.0f, 0.0f};
		float white[] = {1.0f, 1.0f, 1.0f, 0.0f};
		float gray[] = {0.7f, 0.7f, 0.7f, 0.0f};
		float black[] = {0.0f, 0.0f, 0.0f, 0.0f};
		float wood[] = {0.4f, 0.2f, 0.0f, 0.0f};

		// Hills
		if (color) glColor3fv(brown);
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, brown);
		drawHill(8.0f, 0.0f, -8.0f, 2.0f, 1.0f);
		drawHill(0.0f, 0.0f, -9.0f, 1.0f, 2.0f);
		drawHill(-7.0f, 0.0f, -7.0f, 3.0f, 3.5f);
		drawHill(9.0f, 0.0f, -4.0f, 1.0f, 1.5f);
		drawHill(-8.0f, 0.0f, 0.0f, 2.0f, 1.0f);
	
		// Windmill
		if (color) glColor3fv(white);
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);
		glPushMatrix();
			glTranslatef(0.0f, 0.0f, -3.0f);
			glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
			// Base
			glPushMatrix();
				glScalef(-1.0f, 1.0f, 1.0f);
				gluDisk(gQuadric, 0.0f, 2.0f, 8, 1);
			glPopMatrix();
			gluCylinder(gQuadric, 2.0f, 1.5f, 4.0f, 8, 1);
			if (color) glColor3fv(gray);
			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, gray);
			glTranslatef(0.0f, 0.0f, 4.0f);
			// Roof
			gluCylinder(gQuadric, 1.5f, 0.1f, 2.0f, 8, 1);
			glPushMatrix();
				glTranslatef(0.0f, 0.0f, 2.0f);
				gluDisk(gQuadric, 0.0f, 0.1f, 8, 1);
			glPopMatrix();
			gluDisk(gQuadric, 0.0f, 1.5f, 8, 1);
			glTranslatef(0.0f, -2.0f, -1.0f);
			glRotatef(gWindmill, 0.0f, 1.0f, 0.0f);
			glPushMatrix();
				glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
				glScalef(0.1f, 0.1f, 0.5f);
				glTranslatef(0.0f, 0.0f, 0.5f);
				// Rod
				if (color) glColor3fv(black);
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, black);
				glutSolidCube(1.0f);
			glPopMatrix();
			if (color) glColor3fv(wood);
			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, wood);
			for (int i = 0; i < 4; i++) {
				glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
				glPushMatrix();
					glTranslatef(1.0f, 0.0f, 0.0f);
					glScalef(2.0f, 0.1f, 0.1f);
					// Arm
					glutSolidCube(1.0f);
				glPopMatrix();
	
				glPushMatrix();
					//glTranslatef(1.25f, 0.0f, 0.245f);
					glTranslatef(1.25f, -0.0125f, 0.275f);
					glScalef(1.5f, 0.1f, 0.5f);
					// Blade
					glutSolidCube(1.0f);
				glPopMatrix();
			}
		glPopMatrix();
}

void display() {
	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

	glPushMatrix();
		glRotatef(-gCameraRotX, 1.0f, 0.0f, 0.0f);
		glRotatef(-gCameraRot, 0.0f, 1.0f, 0.0f);
	
		float walk[3] = {0.0f, 0.0f, 0.0f};
	
		if (gOptions.movement) {
			float theta = float(gCameraRot * M_PI / 180.0f);
			float waver = 0.1f * cos(gWobble * M_PI / 180.0f);
			walk[0] = waver * cos(theta);
			walk[1] = 0.05f * (1.0f + sin(2 * gWobble * M_PI / 180.0f));
			walk[2] = waver * sin(theta);
		}
	
		glTranslatef(-gCameraPos[0] + walk[0], -gCameraPos[1] + walk[1], -gCameraPos[2] + walk[2]);

		// Draw the ground
		glColor3f(0.0f, 0.5f, 0.0f);
		float green[] = {0.0f, 1.0f, 0.0f, 0.0f};
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green);
		glNormal3f(0.0f, 1.0f, 0.0f);
		glBegin(GL_TRIANGLES);
		glVertex3f(-10.0f, 0.0f, -10.0f);
		glVertex3f(-10.0f, 0.0f, 10.0f);
		glVertex3f(10.0f, 0.0f, -10.0f);
		glVertex3f(10.0f, 0.0f, -10.0f);
		glVertex3f(-10.0f, 0.0f, 10.0f);
		glVertex3f(10.0f, 0.0f, 10.0f);
		glEnd();
		
		glPushAttrib(GL_LIGHTING_BIT|GL_DEPTH_BUFFER_BIT);
		glDisable(GL_LIGHTING);
		glDepthFunc(GL_ALWAYS);
		// Grid lines
		glLineWidth(1.0);
		glColor3f(0.0f, 0.0f, 0.0f);
		glBegin(GL_LINES);
		for (int i = -10; i <= 10; i++) {
			glVertex3f(i, 0.0f, -10.0f);
			glVertex3f(i, 0.0f, 10.0f);
			glVertex3f(-10.0f, 0.0f, i);
			glVertex3f(10.0f, 0.0f, i);
		}
		glEnd();

		glClear(GL_DEPTH_BUFFER_BIT);
	
		// Draw axes
		glLineWidth(2.0);
		glBegin(GL_LINES);
		glColor3f(1.0f, 0.0f, 0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(1.0f, 0.0f, 0.0f);
		glColor3f(0.0f, 1.0f, 0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(0.0f, 1.0f, 0.0f);
		glColor3f(0.0f, 0.0f, 1.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glVertex3f(0.0f, 0.0f, 1.0f);
		glEnd();
		glPopAttrib();

		if (gOptions.lighting) {
			glEnable(GL_LIGHTING);
			glEnable(GL_LIGHT0);
		} else {
			glDisable(GL_LIGHTING);
		}
	
		drawScene(true);

		if (gOptions.outline) {
			glPushAttrib(GL_DEPTH_BUFFER_BIT|GL_ENABLE_BIT|GL_POLYGON_BIT);
			glEnable(GL_CULL_FACE);
			glDisable(GL_LIGHTING);
			glCullFace(GL_FRONT);
			glPolygonMode(GL_BACK, GL_LINE);
			glDepthFunc(GL_LEQUAL);
			glLineWidth(3.0f);
			glColor3f(0.0f, 0.0f, 0.0f);
			drawScene(false);
			glPopAttrib();
		}

	glPopMatrix();
	
	glutSwapBuffers();
}


void reshape(int width, int height) {
	glViewport(0, 0, gWidth = width, gHeight = height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, float(width)/float(height?height:1.0f), 0.5f, 300.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void quit() {
	gluDeleteQuadric(gQuadric);
}

void translateCamera(float dx, float dy, float dz) {
	double theta = gCameraRot * M_PI / 180.0;
	float sint = sin(theta);
	float cost = cos(theta);
	gCameraPos[0] += dz * sint + dx * cost;
	gCameraPos[1] += dy;
	gCameraPos[2] += dz * cost - dx * sint;
}

void movementUpdate() {
	const float velocity = 0.1f;
	const float rotspeed = 5.0f;

	if (gKeys['a']) {
		translateCamera(-velocity, 0.0f, 0.0f);
	}

	if (gKeys['d']) {
		translateCamera(velocity, 0.0f, 0.0f);
	}

	if (gKeys['w']) {
		translateCamera(0.0f, 0.0f, -velocity);
	}

	if (gKeys['s']) {
		translateCamera(0.0f, 0.0f, velocity);
	}

	if (gKeys['.']) {
		gCameraRot -= rotspeed;
		if (gCameraRot < 0.0f) gCameraRot = 360.0f;
	}

	if (gKeys[',']) {
		gCameraRot += rotspeed;
		if (gCameraRot > 360.0f) gCameraRot = 0.0f;
	}

	if (gKeys['r']) {
		gWindmill += 10.0f;
	}

	if (gKeys['w'] || gKeys['s'] && gCameraPos[1] <= 1.0f) {
		gWobble += 10.0f;
		if (gWobble > 360.0f) gWobble = 0.0f;
	}
	
	gVelocityY -= 0.005f;
	if (gCameraPos[1]) gCameraPos[1] += gVelocityY;
	if (gCameraPos[1] < 1.0f) {
		gCameraPos[1] = 1.0f;
		gVelocityY = 0.0f;
	}
}

void keyboard(unsigned char key, int x, int y) {
	gKeys[key] = true;

	switch (key) {
		case 'q':
		case 'Q':
			exit(0);
			break;
		case ' ':
			if (gCameraPos[1] <= 1.0f) {
				gVelocityY = 0.1f;
			}
			break;
		case 'i':
		case 'I':
			reset();
			break;
		case '1':
			gOptions.outline = !gOptions.outline;
			break;
		case '2':
			gOptions.lighting = !gOptions.lighting;
			break;
		case '3':
			gOptions.movement = !gOptions.movement;
			break;
	}

	movementUpdate();
}

void keyboardUp(unsigned char key, int x, int y) {
	gKeys[key] = false;
}

void timer(int t) {
	glutTimerFunc(gDelay, timer, 0);

	if (gOptions.movement) {
		movementUpdate();
	}

	glutPostRedisplay();
}

void mouse(int button, int state, int x, int y) {
	gCursor[0] = x;
	gCursor[1] = y;
	if (button >= 0 && button <= 2) {
		gButtons[button] = state == GLUT_DOWN;
	}
}

void motion(int x, int y) {
	gCameraRot -= (90.0f * (x - gCursor[0])) / gWidth;
	gCameraRotX -= (90.0f * (y - gCursor[1])) / gHeight;
	gCursor[0] = x;
	gCursor[1] = y;
}

int main(int argc, char *argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE|GLUT_DEPTH);
	glutInitWindowSize(gWidth = 640, gHeight = 480);
	glutCreateWindow(argv[0]);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutTimerFunc(gDelay, timer, 0);
	glutReshapeFunc(reshape);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	atexit(quit);
	init();
	glutMainLoop();

	return 0;
}
